keypoll W~
FORGET MULTIFAT.fth
TACHYON
[~
pub MULTIFAT.fth    ." FAT32/16 MultiFile Layer V1.1 131111-0000 " ;
IFNDEF SDCARD.fth
CR ." !!!  This module requires SDCARD.fth !!!          "
!!!
}
{
DESCRIPTION
Provide access to FAT formatted SD cards allowing virtual memory access to the first 4GB of the card.
The driver is primarily written for datalogging applications and for loading source code files from SD but
may be expanded to include full FAT32/16 file access words.

TO DO: backport FAT16 layer etc

MODULE INFORMATION
NAMES:  $585B...7496 for 7227 (0747 bytes added)
CODE:   $0000...3EBB for 8832 (1877 bytes added)
CALLS:  0346 vectors free
RAM:    6560 bytes free

NOTES:
Using DS to define variables rather than the LONG/WORD/BYTE means the storage space is not interrupted with
bytecodes, just purely data, so it's safe to block erase etc.
}

\ Partition table image - copied here
TABLE parts 16 4 * ALLOT        \ Room for 4 entries of 16 bytes

\ These are the offset constants compiled as absolute addresses (for 1st partition)
parts    ORG
1    DS+ \ state            \ 00 = inactive, $80 = active
1    DS+ \ head           \ beginning of paritiion
2    DS+ \ cylsec       \ beginning of partition cylinders/sector
1    DS+ \ partype       \ type of partition
1    DS+ \ ehead            \ end of partition head
2    DS+ \ ecylsec       \ end of partitiion cylinder/sector
4    DS fatptr
4    DS+ \ size


TABLE fat32 $60 ALLOT
fat32    ORG    
3    DS+
8    DS oemname
2    DS byte/sect
1    DS sect/clust
2    DS rsvd
1    DS fats            \ Copies of FAT
4    DS+
1    DS+                \ media
2    DS+
2    DS+                \ sect/trk    
2    DS+                \ heads
4    DS+                \ hidden sectors
4    DS sectors            \ Number of sectors * byte/sect (512) = capacity
4    DS sect/fat
2    DS fatflags
2    DS+                \ fatver
4    DS root            \ Cluster Number of the Start of the Root Directory
2    DS+                \ info = Sector Number of the FileSystem Information Sector  (from part start)
2    DS+                \ boot = Sector Number of the Backup Boot Sector (from part start) 
#12    DS+
3    DS+
4    DS serial            \ serial number of partition
#11    DS volname            \ volume name
8    DS fatname

\ file system variables
LONG    rootdir
BYTE    mounted,clshift
mounted C~

pub @FAT ( index -- sector )
    sect/fat @ * 
     fatptr @ rsvd W@ +            \ look at FAT1 and read sector
     +
    ;
pub CLUST>SECT ( clust# -- sector )
    root @ - clshift C@ SHL rootdir @ +
    ;
pub READFAT32
     fatptr @ SECTOR                \ Read the FAT32 boot record
     SDBUF fat32 $60 CMOVE            \ backup the FAT32 entry for direct access
    sect/clust C@ >| clshift C!        \ make a fast multiply using a shift left constant
     rsvd W@ sect/fat @ fats C@ *   
     root @ 2 - sect/clust C@ * + + fatptr @ + rootdir !
    ;

pub .FAT
     CR ." Media mounted as "
      fatptr @ SECTOR                \ read FAT32 boot record
     cid 9 + U@ .LONG                \ print serial#
     SPACE oemname 8 TYPE            \ print OEM name
     SPACE 
     volname #11 TYPE                \ print volume name
     SPACE
     fatname 8 TYPE                \ print FAT32 name
    ."  Cluster size = " sect/clust C@ BLKSIZ * .DECX ."    Sectors = " sectors @ .DECX  
    ;


\ A copy of the original virtual address of the directory entry is kept in this array (4 entries)
TABLE diradr 4 4 * ALLOT                    \ virtual memory address of file's directory entry

TABLE file 4 4 * ALLOT                    \ table entry for 4 open files

\ Directory entry table
TABLE dirbuf BL 4 * ALLOT                \ create a directory buffer for the opened file

pub ?MOUNT
    CARD? NOT IF ."  *No Card inserted!* " mounted C~ THEN
    mounted C@ IF EXIT THEN
pub MOUNT
     mounted C~ wrflg C~
     sector 16 ERASE
     file 16 ERASE
     dirbuf $80 ERASE
     diradr 16 ERASE                    \ erase dir entry address pointers 
     !SD 
     IF
       wrflg C~ mounted C~~
       0 SECTOR                        \ read partition info
       $01FE SDBUF + W@ $AA55 =
       IF
         $01BE SDBUF + parts $40 CMOVE        \ buffer patitions
        READFAT32                    \ read the FAT32 sector
        ." Mounted SD Card " .FAT            \ display 
      ELSE
          ."  *Format Error* "
      THEN
    ELSE ." *Card Error* "
    THEN
    ;



( DIRECTORY HANDLERS )

pub .FDATE ( fdate -- ; print fat formatted date )
      DUP $1F AND $20A .NUM ." /"
      5 SHR DUP $0F AND $20A .NUM    ." /"
      4 SHR #1980 + $40A .NUM
      ;
pub .FTIME ( ftime -- ; print fat formatted time )
      DUP #11 SHR $20A .NUM ." :"
      DUP 5 SHR $3F AND $20A .NUM ." :"
      $1F AND 2* $20A .NUM
      ;
{
Attribute flags:
0    Read Only
1    Hidden
2    System
3    Volume label 
4    Directory
5    Archive
}

pri .ATR ( atr -- )
    " rhsvda" SWAP 6 FOR DUP 1 AND IF OVER C@ EMIT ELSE ." ." THEN 2/ SWAP 1+ SWAP NEXT 2DROP
    ;
\ Directory entry offsets
#00    == @FNAME
#08    == @FEXT
#11    == @ATR
#13    == @CMS
#14    == @CTIME
#16    == @CDATE
#18    == @ADATE
#20    == @FCLSTH
#22    == @FTIME
#24    == @FDATE
#26    == @FCLST
#28    == @FSIZE

pub CLUSTER@ ( diradr -- cluster )
    DUP @FCLST + W@ SWAP @FCLSTH + W@ W>L
    ;

\ FTP LIST format    -rwxrwxrwx    1 502      500      32768 Mar  4  2013 LOG0001.TXT

" JanFebMarAprMayJunJulAugSepOctNovDec" 0 STRING MONTHS$

pri .DECPAD ( num n -- )
     #10 base C! SWAP <# #S #> DUP LEN$ ROT SWAP - SPACES PRINT$
     ;
pri (.LIST) 
    ." -rwxrwxrwx    1 502      500"
     I @FSIZE + @ 8 .DECPAD SPACE
     I @FDATE + W@ 5 SHR $0F AND 1- 3 * MONTHS$ + 3 TYPE
     I @FDATE + W@ $1F AND 3 .DECPAD
     I @FDATE + W@ 9 SHR #1980 + 6 .DECPAD
     SPACE
     I 8 ADO I C@ BL > IF I C@ EMIT THEN LOOP ." ." I 8 + 3 TYPE
    CR
    ;


pri .DIR ( addr -- )
    DUP 8 TYPE ." ." DUP 8 + 3 TYPE
     SPACE DUP @ATR + C@ .ATR SPACE
    DUP CLUSTER@ CLUST>SECT .LONG SPACE
     3 SPACES DUP @FDATE + W@ .FDATE
     SPACE DUP @FTIME + W@ .FTIME
     3 SPACES @FSIZE + @ .DECX
    ;

pri (.DIR) I CR .DIR ;

BYTE lscnt
pri (ls)
     lscnt C@ 7 AND 0= IF CR THEN                    \ 8 files per line 
     I 8 TYPE ." ." I 8 + 3 TYPE 5 SPACES 1 lscnt C+!
    ;

WORD udir
pub DIR
    ' (.DIR)
pri (DIR) ( code -- )
    udir W!
    ?MOUNT
     TAB volname #11 TYPE CR
    rootdir @ sect/clust C@            \ scan and list rootdir by sector up to one cluster long 
     ADO 
       I SECTOR
       ACTIVE?
       IF
        SDBUF $200                 \ scan the buffer for dir entries
         ADO
           I @ATR + C@ $0F >            \ accept only valid entries
         I C@ $80 < AND            \ but ignore if 1st character has 8th bit set
           I C@ "?" <> AND            \ ignore "deleted" entries
            IF udir W@ CALL THEN        \ all good, list this one
        BL +LOOP
       ELSE
         LEAVE
      THEN
    LOOP
    ;

pub .LIST ' (.LIST) (DIR) ;

\ list the directory in wide and simple format

pub ls
' (ls) (DIR) 
    ;




" FILENAMETXT" 0 STRING dir$

pub DIR?    ( str -- diraddr | FALSE )
    ?MOUNT
      mounted C@ IF
      >L                             \ save str ptr onto loop stack - ref as IX
       rootdir @ 9 SHL                 \ get virtual address of root directory
       BEGIN ( dirptr )
         DUP XC@                        \ assume the file is in the root cluster
       WHILE
        DUP XADR dir$ #11 CMOVE dir$        \ copy the name as a string
        IX COMPARE$ IF L> DROP EXIT THEN    \ Found it, exit here with address
         BL +                        \ skip to next dir entry (32 bytes)
       REPEAT
       L> 2DROP                        \ failed to find it - discard parameters
     ELSE
        DROP
     THEN
     FALSE                            \ Failed to find it
      ;



pub @DIR ( -- bufadr )
     filesel C@ 5 SHL dirbuf +            \ index 1 of 4 directory buffers
     ;


" FILENAMESTR" 0 STRING fname$

\ format friendly file name into directory format  
pri >F83 ( str1 -- str2 )
    fname$ #11 BL FILL
    DUP fname$ ( str1 str1 fname$ )
    "." 3RD LOCATE$ ( str1 str1 fname$ ext )
    ?DUP IF
       3RD - CMOVE
       "." SWAP LOCATE$ 1+ fname$ 8 + 3 CMOVE
      ELSE ROT STRLEN CMOVE THEN
    fname$
    ;

pub FSIZE@        @DIR @FSIZE + @ ;

\ Fetch the address of the currently open file 
pub FILE@ (  -- xadr )    file @FILE @ 9 SHL ;

pub FCLOSE
    FLUSH
     sector @FILE ~
     file @FILE ~
     ;
\ Open the file with the formatted name and return with it's sector (0=fail)
pub FOPEN$ ( namestr -- sector )
     FCLOSE
    >F83 DIR? DUP 
     IF 
       DUP diradr @FILE !             \ save address of directory entry
       XADR DUP @DIR BL CMOVE         \ copy directory entry to local buffer
       CLUSTER@ CLUST>SECT     
       DUP file @FILE !             \ save address of file 
      THEN
    ;


IFNDEF WRSECT
pub WRSECT ( write the current sector back to the storage media )
     SDBUF sector @FILE @ SDWR
    ;
}

pub RENAME$ ( new$ -- )
FILE@ IF 
  >F83 @DIR #11 CMOVE                 \ update local copy of directory entry
  @DIR diradr @FILE @ XADR #11 CMOVE             \ copy that to the virtual address
 WRSECT                            \ Force an update
     THEN
     DROP
    ;
    
    

\ Open a file with the name specified in the input stream
pub FOPEN ( <name> -- )
    GETWORD
     FOPEN$ ?DUP 
     IF ." ...opened at "  .LONG 
     ELSE ."  not found " 
     THEN
    ;
IMMEDIATE

LONG xadr    \ virtual memory address pointer - used in file load

\ Read in the next character in the virtual buffer as part of the console input stream
pri GETCH
    xadr @ XC@ 1 xadr +!            \ Read a single character 
    DUP 0= IF CON THEN            \ stop when null encountered
    DUP $0A = IF DROP $0D THEN        \ Replace LFs with CRs
    ;

\ Load a file as console input - scripts or source code
pub FLOAD ( <name> -- )
    GETWORD >F83
    FOPEN$ ?DUP
    IF 
       9 SHL xadr !             \ Set "xadr" as the load pointer
       ' GETCH ukey W!            \ and redirect console input to this virtual buffer
     THEN
    ;
IMMEDIATE

pub cat ( <name> -- ) \ List the file as text 
    [COMPILE] FOPEN
    CR FILE@ BEGIN DUP XC@ ?DUP WHILE DUP $0A = IF $0D EMIT THEN EMIT 1+ REPEAT DROP
    ;
IMMEDIATE

pub FILES
     4 0 DO CR ." #" I . SPACE I FILE @DIR .DIR LOOP    
     ;

pub SAVEROM 
    " ROM.BIN" FOPEN$ IF ."  SAVING TO ROM.BIN " 0 FILE@ $8000 XSAVE THEN
    ;
pub LOADROM
    " ROM.BIN" FOPEN$ IF ." Loading ROM to EEPROM " THEN
     
    ;


]~
END
?BACKUP


